/*
 * This builds Windows compiled html help compatable files
 * from the CGNS html documentation. The program is not as robust
 * as it could be.
 *
 * The navigation headers are striped from the html files
 * since this is provided by the index navigation pane.
 *
 * Windows seems to have problems with the ':' in topic names
 * when used in the navigation pane, so these are all replaced
 * with a '-' in the html files.
 *
 * The .hhp file is built from the list of html files,
 * and the sitemaps directory is scanned to build the
 * .hhc index file.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/types.h>
#ifdef _WIN32
#include <direct.h>
#include "dirent.h"
#define STRCMP  stricmp
#define STRNCMP strnicmp
#else
#include <unistd.h>
#include <dirent.h>
#define STRCMP  strcasecmp
#define STRNCMP strncasecmp
#endif

#define CHMNAME    "cgns"
#define SUBCHAR    '-'
#define MAXPATHLEN 1024
#define CGNSHOME   "http://www.grc.nasa.gov/WWW/cgns"

#define ADD_EXTERN

static int all = 0;
static int verbose = 0;
static int level = 0;
static char curdir[MAXPATHLEN+1];
static char inpdir[MAXPATHLEN+1];
static char outdir[MAXPATHLEN+1];
static char cgnshome[MAXPATHLEN+1];

static char buff[1024];

static FILE *hhfp;

static const char *skip_dir[] = {
    "CGNSTalk",
    "minutes",
    "papers",
    "slides"
};
#define NUM_SKIP_DIR 4

static const char *skip_ext[] = {
    ".gz",
    ".pdf",
    ".htm"
};
#define NUM_SKIP_EXT 3

/*-------------------------------------------------------------------*/

static void create_directory (char *dirname)
{
    struct stat st;

    if (stat (dirname, &st)) {
#ifdef _WIN32
        if (mkdir (dirname)) {
#else
        if (mkdir (dirname, 0777)) {
#endif
            fprintf (stderr, "couldn't create directory <%s>\n", dirname);
            exit (1);
        }
    }
    else {
        if ((st.st_mode & S_IFDIR) != S_IFDIR) {
            fprintf (stderr, "<%s> exists and is not a directory\n", dirname);
            exit (1);
        }
    }
}

/*-------------------------------------------------------------------*/

static void indent ()
{
    int n;

    for (n = 0; n < level; n++)
        putchar (' ');
}

/*-------------------------------------------------------------------*/

static char *next_reference (FILE *finp, FILE *fout, int skip)
{
    int c, n;

    while ((c = getc (finp)) != EOF) {
        /* ignore CR on input */
        if (c == '\r') continue;
        /* write CR/LF when LF found */
        if (c == '\n') {
            if (!skip) putc ('\n', fout);
            continue;
        }
        if (c == '<') {
            while ((c = getc (finp)) != EOF && isspace(c))
                ;
            if (c == EOF) return NULL;
            if (c != 'a' && c != 'A') {
                if (!skip) {
                    putc ('<', fout);
                    putc (c, fout);
                }
                continue;
            }
            strcpy (buff, "<a");
            n = 2;
            while ((c = getc(finp)) != EOF && c != '>') {
                if (n == sizeof(buff)) {
                    fprintf (stderr, "exceeded buff in next_reference\n");
                    exit (1);
                }
                if (isspace (c)) {
                    if (buff[n-1] != ' ') buff[n++] = ' ';
                }
                else
                    buff[n++] = c;
            }
            if (c == EOF) return NULL;
            if ((n + 2) >= sizeof(buff)) {
                fprintf (stderr, "exceeded buff in next_reference\n");
                exit (1);
            }
            buff[n++] = '>';
            buff[n] = 0;
            return buff;
        }
        if (!skip) putc (c, fout);
    }
    return NULL;
}

/*-------------------------------------------------------------------*/

#ifdef ADD_EXTERN

static void end_reference (FILE *finp, FILE *fout)
{
    int c;

    while ((c = getc (finp)) != EOF) {
        if (c == '\r') continue;
        putc(c, fout);
        if (c != '<') continue;
        if ((c = getc (finp)) == EOF) return;
        putc(c, fout);
        if (c != '/') continue;
        if ((c = getc (finp)) == EOF) return;
        putc(c, fout);
        if (c != 'a' && c != 'A') continue;
        if ((c = getc (finp)) == EOF) return;
        while (isspace(c)) {
            putc(c, fout);
            if ((c = getc (finp)) == EOF) return;
        }
        putc(c, fout);
        if (c == '>') return;
    }
}

#endif

/*-------------------------------------------------------------------*/

static int check_external(char *href)
{
    char *p, fname[MAXPATHLEN+1];
    int n;

    if (strlen(href) > MAXPATHLEN) {
        fprintf(stderr, "exceeded MAXPATHLEN in check_external\n");
        exit(1);
    }
    strcpy(fname, href);
    if ((p = strchr (fname, '#')) != NULL) *p = 0;
    if (*fname && access(fname, 0)) return 1;
    if (all) return 0;

    for (n = 0; n < NUM_SKIP_DIR; n++) {
        p = strstr(fname, skip_dir[n]);
        if (p != NULL) {
            p += strlen(skip_dir[n]);
            if (*p == '/') return 1;
        }
    }
    p = strrchr(fname, '.');
    if (p != NULL) {
        for (n = 0; n < NUM_SKIP_EXT; n++) {
            if (0 == STRCMP(p, skip_ext[n])) return 1;
        }
    }
    return 0;
}

/*-------------------------------------------------------------------*/

static void fix_html (char *dirname, char *htmlfile)
{
    FILE *finp, *fout;
    int n;
    char *p, *s, *ps, *fn;
    char *loc = outdir + strlen(outdir);

    if (dirname == NULL) {
        fprintf (hhfp, "%s\n", htmlfile);
        sprintf (loc, "/%s", htmlfile);
    }
    else {
        fprintf (hhfp, "%s/%s\n", dirname, htmlfile);
        sprintf (loc, "/%s/%s", dirname, htmlfile);
    }

    if (verbose) {
        indent ();
        printf ("%s\n", htmlfile);
    }
    unlink (outdir);

    if ((finp = fopen (htmlfile, "rb")) == NULL) {
        fprintf (stderr, "couldn't open <%s> for reading\n", htmlfile);
        exit (1);
    }
    if ((fout = fopen (outdir, "w+b")) == NULL) {
        fprintf (stderr, "couldn't open <%s> for writing\n", outdir);
        exit (1);
    }

    while ((p = next_reference (finp, fout, 0)) != NULL) {

        /* remove navigation header */

        if (0 == STRCMP (p, "<a href=\"#endnav\">")) {
            while ((p = next_reference (finp, fout, 1)) != NULL) {
                if (0 == STRCMP (p, "<a name=\"endnav\">"))
                    break;
            }
            if (verbose) {
                indent();
                printf (" removed navigation header\n");
            }
            continue;
        }

        if (0 == STRNCMP (p, "<a href=\"http:", 14) ||
            0 == STRNCMP (p, "<a href=\"https:", 15) ||
            0 == STRNCMP (p, "<a href=\"mailto:", 16)) {
            fputs (p, fout);
            continue;
        }

        /* if href file does not exist, point it to grc */

        if (0 == STRNCMP (p, "<a href=\"", 9)) {
            fn = p + 9;
            if ((s = strchr (fn, '"')) == NULL) {
                fprintf (stderr, "missing \" in fix_html for href\n");
                exit (1);
            }
            *s = 0;
            if (check_external(fn)) {
                if (verbose) {
                    indent();
                    printf (" changed href <%s> -> ", fn);
                }
                if (dirname == NULL) {
                    if (verbose)
                        printf ("<%s/%s>\n", cgnshome, fn);
                    fprintf (fout, "<a href=\"%s/%s\">", cgnshome, fn);
                }
                else {
                    if (verbose)
                        printf ("<%s/%s/%s>\n", cgnshome, dirname, fn);
                    fprintf (fout, "<a href=\"%s/%s/%s\">",
                        cgnshome, dirname, fn);
                }
#ifdef ADD_EXTERN
                end_reference(finp, fout);
                fprintf(fout,
                    " <img src=\"external.png\" alt=\"external link\">");
#endif
                continue;
            }

            ps = strchr(fn, '#');
            if (ps != NULL && (s = strchr (ps, ':')) != NULL) {
                if (verbose) {
                    indent();
                    printf (" fixed href <%s> -> ", fn);
                }
                while (*s) {
                    if (*s == ':') *s = SUBCHAR;
                    s++;
                }
                if (verbose)
                    printf ("<%s>\n", fn);
            }

            fprintf (fout, "<a href=\"%s\">", fn);
            continue;
        }

        if (0 == STRNCMP (p, "<a name=\"", 9) &&
            (s = strchr (p + 9, ':')) != NULL) {
            fn = p + 9;
            if ((ps = strchr (fn, '"')) == NULL) {
                fprintf (stderr, "missing \" in fix_html for name\n");
                exit (1);
            }
            *ps = 0;
            if (verbose) {
                indent();
                printf (" fixed name <%s> -> ", fn);
            }
            while (*s && *s != '"') {
                if (*s == ':') *s = SUBCHAR;
                s++;
            }
            if (verbose)
                printf ("<%s>\n", fn);
            fprintf (fout, "<a name=\"%s\">", fn);
            continue;
        }

        fputs (p, fout);
    }

    fclose (finp);
    fclose (fout);
    *loc = 0;
}

/*-------------------------------------------------------------------*/

static void copy_file (char *dirname, char *fname)
{
    int c;
    FILE *finp, *fout;
    char *loc = outdir + strlen(outdir);

    if (dirname == NULL)
        sprintf (loc, "/%s", fname);
    else
        sprintf (loc, "/%s/%s", dirname, fname);

    if (verbose) {
        indent ();
        printf ("%s\n", fname);
    }
    unlink (outdir);

    if ((finp = fopen (fname, "rb")) == NULL) {
        fprintf (stderr, "couldn't open <%s> for reading\n", fname);
        exit (1);
    }
    if ((fout = fopen (outdir, "w+b")) == NULL) {
        fprintf (stderr, "couldn't open <%s> for writing\n", outdir);
        fclose (finp);
        exit (1);
    }
    while ((c = getc (finp)) != EOF)
        putc (c, fout);

    fclose (finp);
    fclose (fout);
    *loc = 0;
}

/*-------------------------------------------------------------------*/

static void parse_dir (char *dirname)
{
    DIR *dirp;
    struct dirent *dp;
    char *p;
    struct stat st;
    int n;

    if (dirname != NULL) {
        if (verbose) {
            indent ();
            printf ("[%s]\n", dirname);
        }
        p = outdir + strlen(outdir);
        sprintf (p, "/%s", dirname);
        create_directory (outdir);
        *p = 0;
    }
    if ((dirp = opendir (".")) == NULL) {
        fprintf (stderr, "error opening current directory\n");
        exit (1);
    }
    level++;

    while ((dp = readdir (dirp)) != NULL) {
        if (*dp->d_name == '.') continue;
        if (stat (dp->d_name, &st)) {
            fprintf (stderr, "couldn't stat <%s>\n", dp->d_name);
            exit (1);
        }
        if ((st.st_mode & S_IFDIR) == S_IFDIR) continue;
        p = strrchr (dp->d_name, '.');
        if (p != NULL && 0 == STRCMP (p, ".html")) {
            fix_html (dirname, dp->d_name);
        }
        else if (!all && p != NULL) {
            for (n = 0; n < NUM_SKIP_EXT; n++) {
                if (0 == STRCMP (p, skip_ext[n])) break;
            }
            if (n == NUM_SKIP_EXT)
                copy_file (dirname, dp->d_name);
        }
        else {
            copy_file (dirname, dp->d_name);
        }
    }

    rewinddir (dirp);
    while ((dp = readdir (dirp)) != NULL) {
        if (*dp->d_name == '.' ||
            0 == strcmp (dp->d_name, "sitemaps")) continue;
        if (!all) {
            for (n = 0; n < NUM_SKIP_DIR; n++) {
                if (0 == strcmp (dp->d_name, skip_dir[n])) break;
            }
            if (n < NUM_SKIP_DIR) continue;
        }
        if (stat (dp->d_name, &st)) {
            fprintf (stderr, "couldn't stat <%s>\n", dp->d_name);
            exit (1);
        }
        if ((st.st_mode & S_IFDIR) == S_IFDIR) {
            if (dirname == NULL) {
                p = inpdir;
                strcpy (p, dp->d_name);
            }
            else {
                p = inpdir + strlen(inpdir);
                sprintf (p, "/%s", dp->d_name);
            }
            if (chdir (dp->d_name)) {
                fprintf (stderr, "couldn't chdir to <%s>\n", dp->d_name);
                exit (1);
            }
            parse_dir (inpdir);
            chdir ("..");
            *p = 0;
        }
    }

    level--;
}

/*-------------------------------------------------------------------*/

static char *next_index (FILE *fp, int need_li, int *nl)
{
    char *p, *s;
    int c, n, found_li = 0;

    *nl = 0;
    while ((c = getc (fp)) != EOF) {
        if (c != '<') continue;
        buff[0] = c;
        n = 1;
        while ((c = getc(fp)) != EOF && c != '>') {
            if (n == sizeof(buff)) {
                fprintf (stderr, "exceeded buff in next_index\n");
                exit (1);
            }
            if (c == '\n' || c == '\r' || c == '\t')
                buff[n++] = ' ';
            else
                buff[n++] = c;
        }
        if (c == EOF) return NULL;
        if ((n + 2) >= sizeof(buff)) {
            fprintf (stderr, "exceeded buff in next_index\n");
            exit (1);
        }
        buff[n++] = '>';
        buff[n] = 0;

        if (0 == STRCMP (buff, "<ul>")) {
            (*nl)++;
            continue;
        }
        if (0 == STRCMP (buff, "</ul>")) {
            (*nl)--;
            continue;
        }
        if (0 == STRCMP (buff, "<li>")) {
            found_li = 1;
            continue;
        }
        if (STRNCMP (buff, "<a href=\"", 9)) {
            found_li = 0;
            continue;
        }
        if (need_li && !found_li) continue;

        p = buff + 9;
        n = 0;
        while (*p && *p != '"' && *p != '>') {
            buff[n++] = *p++;
        }
        buff[n++] = 0;

        while ((c = getc (fp)) != EOF) {
            if (c == '<') {
                if ((c = getc (fp)) == EOF) return NULL;
                if (c == '/') {
                    if ((c = getc (fp)) == EOF) return NULL;
                    if (c == 'a' || c == 'A') {
                        buff[n] = 0;
                        return buff;
                    }
                }
                do {
                    if ((c = getc (fp)) == EOF) return NULL;
                } while (c != '>');
                continue;
            }
            if (n == sizeof(buff)) {
                fprintf (stderr, "exceeded buff in next_index\n");
                exit (1);
            }
            if (isspace (c)) {
                if (buff[n-1] != ' ') buff[n++] = ' ';
            }
            else
                buff[n++] = c;
        }
    }
    return NULL;
}

/*-------------------------------------------------------------------*/

static void add_index (char *fname)
{
    char *p, *s;
    int n, nl;
    FILE *fp;

    if ((fp = fopen (fname, "rb")) == NULL) return;
    if (verbose)
        printf("reading index from <%s>\n", fname);

    /* skip navigation bar */

    do {
        p = next_reference (fp, NULL, 1);
        if (p == NULL) {
            fclose (fp);
            return;
        }
    } while (STRCMP (p, "<a name=\"endnav\">"));

    /* first index entry */

    if ((p = next_index (fp, 0, &nl)) == NULL) {
        fclose (fp);
        return;
    }
    if (0 == strncmp (p, "../", 3)) p += 3;
    s = p + strlen (p) + 1;
    if (verbose)
        printf (" %s <%s>\n", s, p);
    fprintf (hhfp, "<LI> <OBJECT type=\"text/sitemap\">\n");
    fprintf (hhfp, "<param name=\"Name\" value=\"%s\">\n", s);
    fprintf (hhfp, "<param name=\"Local\" value=\"%s\">\n", p);
    fprintf (hhfp, "</OBJECT>\n");

    level = 0;

    /* get index entries */

    while ((p = next_index (fp, 1, &nl)) != NULL) {
        if (nl > 0) {
            for (n = 0; n < nl; n++) {
                fprintf (hhfp, "<UL>");
                level++;
            }
            putc ('\n', hhfp);
        }
        if (nl < 0) {
            for (n = 0; n < -nl; n++) {
                fprintf (hhfp, "</UL>");
                level--;
            }
            putc ('\n', hhfp);
        }
        if (0 == strncmp (p, "../", 3)) p += 3;
        if ((s = strchr (p, '#')) != NULL) {
            while (*++s) {
                if (*s == ':') *s = SUBCHAR;
            }
        }
        s = p + strlen (p) + 1;
        if (verbose) {
            indent ();
            printf (" %s <%s>\n", s, p);
        }
        fprintf (hhfp, "<LI> <OBJECT type=\"text/sitemap\">\n");
        fprintf (hhfp, "<param name=\"Name\" value=\"%s\">\n", s);
        fprintf (hhfp, "<param name=\"Local\" value=\"%s\">\n", p);
        fprintf (hhfp, "</OBJECT>\n");
    }
    if (level > 0) {
        for (n = 0; n < level; n++)
            fprintf (hhfp, "</UL>");
        putc ('\n', hhfp);
    }

    fclose (fp);
}

/*-------------------------------------------------------------------*/

static void build_index ()
{
    FILE *fp;
    int nl;
    char *p, *s;

    if ((fp = fopen ("index.html", "rb")) == NULL) {
        fprintf (stderr, "couldn't open <index.html> for reading\n");
        return;
    }

    while ((p = next_index (fp, 1, &nl)) != NULL) {
        if (access (p, 0) == 0)
            add_index (p);
    }

    fclose (fp);
}

/*-------------------------------------------------------------------*/

int main (int argc, char *argv[])
{
    int n;
    char *p;
    int sitemaps = 0;
    struct stat st;

    if (argc < 3) {
        printf ("usage: html2chm [-a] [-v] input_dir output_dir [CGNShome]\n");
        printf ("arguments:\n");
        printf ("-a = include all (CGNSTalk, minutes, slides, .etc)\n");
        printf ("-v = verbose output\n");
        printf ("input_dir = directory containing CGNS documentation\n");
        printf ("output_dir = directory to write CHM output\n");
        printf ("CGNShome = GRC documentation home for external links\n");
        printf ("           default is CGNS_docs_current\n");
        return 0;
    }
    if (NULL == getcwd (curdir, MAXPATHLEN)) {
        fprintf (stderr, "getcwd failed\n");
        return 1;
    }
    for (n = 1; n < argc && argv[n][0] == '-'; n++) {
        switch (argv[n][1]) {
            case 'a':
                all = 1;
                break;
            case 'v':
                verbose = 1;
                break;
            default:
                fprintf(stderr, "unknown option %s\n", argv[n]);
                return 1;
        }
    }
    if (n + 1 >= argc) {
        fprintf(stderr, "input and output directory not given\n");
        return 1;
    }

    /* input directory */

    if (stat (argv[n], &st)) {
        fprintf (stderr, "couldn't stat <%s>\n", argv[n]);
        return 1;
    }
    if ((st.st_mode & S_IFDIR) != S_IFDIR) {
        fprintf (stderr, "<%s> is not a directory\n", argv[n]);
        return 1;
    }
    if (chdir (argv[n])) {
        chdir (curdir);
        fprintf (stderr, "couldn't cd to <%s>\n", argv[n]);
        return 1;
    }
    if (NULL == getcwd (inpdir, MAXPATHLEN)) {
        chdir (curdir);
        fprintf (stderr, "getcwd failed\n");
        return 1;
    }
    if (access ("index.html", 0)) {
        chdir (curdir);
        fprintf (stderr, "index.html does not exist in <%s>\n", inpdir);
        return 1;
    }
    if (stat ("sitemaps", &st) == 0 &&
        (st.st_mode & S_IFDIR) == S_IFDIR &&
        access("sitemaps/index.html", 0) == 0) {
        sitemaps = 1;
    }
    if (chdir (curdir)) {
        fprintf (stderr, "couldn't cd to <%s>\n", curdir);
        return 1;
    }
    for (p = inpdir; *p; p++) {
        if (*p == '\\') *p = '/';
    }

    /* output directory */

    create_directory (argv[++n]);
    if (chdir (argv[n])) {
        chdir (curdir);
        fprintf (stderr, "couldn't cd to <%s>\n", argv[n]);
        return 1;
    }
    if (NULL == getcwd (outdir, MAXPATHLEN)) {
        chdir (curdir);
        fprintf (stderr, "getcwd failed\n");
        return 1;
    }
    chdir (curdir);
    for (p = outdir; *p; p++) {
        if (*p == '\\') *p = '/';
    }

    /* get CGNS doc home if given */

    strcpy(cgnshome, CGNSHOME);
    if (++n < argc) {
        if (0 == strncmp(argv[n], "http://", 7)) {
            strcpy(cgnshome, argv[n]);
        }
        else {
            p = cgnshome + strlen(cgnshome);
            if (argv[n][0] != '/') *p++ = '/';
            strcpy(p, argv[n]);
        }
    }
    else {
        strcat(cgnshome, "/CGNS_docs_current");
    }

    printf ("input directory  : %s\n", inpdir);
    printf ("output directory : %s\n", outdir);
    printf ("CGNS doc home    : %s\n", cgnshome);

    if (0 == strcmp (inpdir, outdir)) {
        fprintf (stderr, "input and output directories are the same\n");
        return 1;
    }

    /* create winhelp hhp file */

    p = outdir + strlen(outdir);
    sprintf (p, "/%s.hhp", CHMNAME);
    if ((hhfp = fopen (outdir, "w+")) == NULL) {
        fprintf (stderr, "couldn't open <%s> for writing\n", outdir);
        return 1;
    }
    printf ("\ncreating HTML help file  : %s\n", outdir);
    *p = 0;

    fprintf (hhfp, "[OPTIONS]\n");
    fprintf (hhfp, "Binary TOC=Yes\n");
    fprintf (hhfp, "Compatibility=1.1 or later\n");
    fprintf (hhfp, "Compiled file=%s.chm\n", CHMNAME);
    fprintf (hhfp, "Contents file=%s.hhc\n", CHMNAME);
    fprintf (hhfp, "Default Window=default\n");
    fprintf (hhfp, "Default topic=index.html\n");
    fprintf (hhfp, "Display compile progress=No\n");
    fprintf (hhfp, "Full-text search=Yes\n");
    fprintf (hhfp, "Language=0x409 English (United States)\n");
    fprintf (hhfp, "Title=CGNS Documentation\n");
    fprintf (hhfp, "\n[WINDOWS]\n");
    fprintf (hhfp, "default=\"CGNS Documentation\",");
    if (sitemaps)
        fprintf (hhfp, "\"%s.hhc\"", CHMNAME);
    fprintf (hhfp, ",,\"index.html\",\"index.html\",,,,,");
    fprintf (hhfp, "0x43520,,0x207e,,,,,,,,0\n");
    fprintf (hhfp, "\n[FILES]\n");

    chdir (inpdir);
    if (verbose) printf ("[.]\n");
    parse_dir (NULL);
    if (verbose) putchar('\n');

    fprintf (hhfp, "\n[INFOTYPES]\n");
    fclose (hhfp);

    /* create .hhc index file from sitemaps */

    if (sitemaps) {
        if (chdir ("sitemaps")) {
            chdir (curdir);
            fprintf (stderr, "couldn't chdir to <sitemaps>\n");
            exit (1);
        }
        p = outdir + strlen(outdir);
        sprintf (p, "/%s.hhc", CHMNAME);
        if ((hhfp = fopen (outdir, "w+")) == NULL) {
            fprintf (stderr, "couldn't open <%s> for writing\n", outdir);
            exit (1);
        }
        printf ("creating HTML index file : %s\n", outdir);

        fputs ("<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML//EN\">\n", hhfp);
        fputs ("<HTML>\n<HEAD>\n", hhfp);
        fputs ("<meta name=\"GENERATOR\"", hhfp);
        fputs (" content=\"Microsoft&reg; HTML Help Workshop 4.1\">\n", hhfp);
        fputs ("<!-- Sitemap 1.0 -->\n", hhfp);
        fputs ("</HEAD><BODY>\n<UL>\n", hhfp);
        fputs ("<LI><OBJECT type=\"text/sitemap\">\n", hhfp);
        fputs ("<param name=\"Name\" value=\"CFD General Notation System\">\n", hhfp);
        fputs ("<param name=\"Local\" value=\"index.html\">\n", hhfp);
        fputs ("</OBJECT>\n", hhfp);
        fputs ("<UL>\n", hhfp);
        fputs ("<LI> <OBJECT type=\"text/sitemap\">\n", hhfp);
        fputs ("<param name=\"Name\" value=\"What's New\">\n", hhfp);
        fputs ("<param name=\"Local\" value=\"whatsnew.html\">\n", hhfp);
        fputs ("</OBJECT>\n", hhfp);
        fputs ("<LI> <OBJECT type=\"text/sitemap\">\n", hhfp);
        fputs ("<param name=\"Name\" value=\"Document Versions\">\n", hhfp);
        fputs ("<param name=\"Local\" value=\"changes.html\">\n", hhfp);
        fputs ("</OBJECT>\n", hhfp);
        fputs ("<LI> <OBJECT type=\"text/sitemap\">\n", hhfp);
        fputs ("<param name=\"Name\" value=\"Data Structures Summary\">\n", hhfp);
        fputs ("<param name=\"Local\" value=\"structures.html\">\n", hhfp);
        fputs ("</OBJECT>\n", hhfp);
        fputs ("</UL>\n", hhfp);

        build_index ();

        fputs ("</UL>\n</BODY></HTML>\n", hhfp);
        fclose (hhfp);
    }

    chdir (curdir);
    return 0;
}

